package games.sudoku;

import utilities.StatisticalSummary;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 24-Jan-2006
 * Time: 14:53:21
 * To change this template use File | Settings | File Templates.
 */
public class Correlate {

    final static boolean swapMetric = false;

    final static int numberOfCases = 1000;
    final static int maxDistanceToMeasure = (swapMetric ? 40 : 51);
    final static int[][] data = new int[numberOfCases][maxDistanceToMeasure];

    public static void main (String[] args) throws Exception {
        //howAreWeDoing ();
        int[] iGrid = Sudoku.readGrid ("grids/easy1.txt");
        //Initializer initializer = new NaiveInitializer ();
        Initializer  initializer = new RowCompletionInitializer ();
        //Mutator mutator = new NaiveMutator ();
        Mutator mutator = new RowSwapMutator ();
        boolean[] fixed = Sudoku.getFixedPositions (iGrid);
        initializer.initializeGrid (fixed, iGrid, 3);
        int[] grid = new int[iGrid.length];
        RowPermuter rowPermuter = new RowPermuter ();
        FitnessFunction fitnessFunction = new SimonsFitnessFunction ();
        for (int whichCase = 0; whichCase < numberOfCases; whichCase++) {
            rowPermuter.permute (fixed, iGrid, 9);
            System.arraycopy (iGrid, 0, grid, 0, iGrid.length);
            int lastDistance = 0;
            data[whichCase][0] = (int) fitnessFunction.evaluate (grid, 3);
            while (lastDistance < maxDistanceToMeasure) {

                mutator.mutate (fixed, grid, 3);
                int newDistance = (swapMetric ? Sudoku.swapDistance (grid, iGrid, 9) :
                        Sudoku.hammingDistance (grid, iGrid));
                // Oh oh , be wary, don't select on distance!
                //if (newDistance > lastDistance ) {
                    int fitness = (int) fitnessFunction.evaluate (grid, 3);
                    if (newDistance > 0) {
                        data[whichCase][newDistance - 1] = fitness;
                    }
                    //System.out.println (data[whichCase][newDistance - 1]);
                    lastDistance = newDistance;
                    //System.out.println ("Distance" + newDistance + " fitness " + fitness);
                //}
            }
            System.out.print (".");
            //System.out.print (whichCase + " ");

        }
        System.out.println ();

        double[] correlogram = doCorrelations (data);
        for (int i = 0; i < correlogram.length; i++) {
            System.out.println (i + " : " + correlogram[i]);
        }

    }

    public static double[] doCorrelations (int[][] data) {
        double[][] normalizedData = new double[data.length][data[0].length];
        // for each distance
        for (int distance = 0; distance < data[0].length; distance++) {
            // compute mean and standard deviation
            StatisticalSummary ss = new StatisticalSummary ();
            // add all cases
            for (int whichCase = 0; whichCase < data.length; whichCase++) {
                if (data[whichCase][distance] != 0) {
                    ss.add (data[whichCase][distance]);
                }
                //System.out.println ("Case: " + whichCase + " dist " + distance + " : " + data[whichCase][distance]);
            }
            double mean = ss.mean ();
            double sd = ss.sd ();

            System.out.println ("Distance " + distance + " mean " + mean + " sd " + sd) ;
            // create normalized data
            for (int whichCase = 0; whichCase < data.length; whichCase++) {
                if (sd == 0) {
                    System.out.println (whichCase + " " + distance + " " + sd);
                }
                normalizedData[whichCase][distance] = (data[whichCase][distance] - mean) / sd;
                //System.out.println (" " + normalizedData[whichCase][distance]);
            }
        }

        // normalization done

        //System.exit (0);
        double[] correlogram = new double[maxDistanceToMeasure];
        //int numberOfCrossProducts = numberOfCases * numberOfCases;


        //int distance = 0;
       for (int distance = 0; distance < data[0].length; distance++) {
           int numberOfActualCases = 0;
            correlogram[distance] = 0;
            // for all cases...
            for (int i = 0; i < data.length; i++) {
                if (data[i][distance] != 0) {
                    numberOfActualCases++;
                    correlogram[distance] += normalizedData[i][distance] * normalizedData[i][0];
                }
            }
            correlogram[distance] = correlogram[distance] / numberOfActualCases;
            System.out.println ("Actual cases: " + numberOfActualCases);
        }
        System.out.println ();
        return correlogram;
    }


    public static void howAreWeDoing () throws Exception {
        int[] iGrid = Sudoku.readGrid ("grids/easy1.txt");
        //int[] iGrid = new int[81];
        //Initializer initializer = new NaiveInitializer ();
        Initializer initializer = new RowCompletionInitializer ();
        //Mutator mutator = new NaiveMutator ();
        Mutator mutator = new RowSwapMutator ();
        boolean[] fixed = Sudoku.getFixedPositions (iGrid);
        initializer.initializeGrid (fixed, iGrid, 3);
        int[] grid = new int[iGrid.length];
        System.arraycopy (iGrid, 0, grid, 0, iGrid.length);
        int maxDistanceSoFar = 0;
        while (true) {
            mutator.mutate (fixed, grid, 3);
            //int newDistance = Sudoku.hammingDistance (grid, iGrid);
                int newDistance = (swapMetric ? Sudoku.swapDistance (grid, iGrid, 9) :
                        Sudoku.hammingDistance (grid, iGrid));
            if (newDistance > maxDistanceSoFar) {
                maxDistanceSoFar = newDistance;
                System.out.println ("Max: " + maxDistanceSoFar);
            }
        }
    }

}
