package games.sudoku;

import utilities.StatisticalSummary;

import java.io.BufferedWriter;
import java.io.FileWriter;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 25-Nov-2005
 * Time: 19:19:58
 * To change this template use File | Settings | File Templates.
 */
public class DoStats {

    final static int maxNumberOfEvaluationsWithoutProgress = 100000;
    final static int numberOfTestsPerConfiguration = 30;
    final static Crossover[] hammingCrossovers = {new NoCrossover (), new UniformCrossover (),
            new TwoPointCrossover (), new RowWiseCrossover ()};
    final static Mutator[] hammingMutators = {new NonMutator (), new NaiveMutator (), new UniformSwapMutator (),
            new ExponentialProbabilityNaiveMutator ()};
    final static Crossover[] swapCrossovers = {new NoCrossover (), new RowWiseCrossover (), new PMXCrossover (),
            new UniformSwapCrossover (), new OneCycleCrossover (),
            new MultiCycleCrossover ()};
    final static Mutator[] swapMutators = {new NonMutator (), new RowSwapMutator (),
            new ExponentialProbabilityRowSwapMutator ()};
    final static double[] mutationRates = {0, 0.2, 0.4, 0.6, 0.8};
    final static int[] populationSizes = {100, 500, 1000, 3000, 5000};
    final static double[] proportionsSteadyState = {0.01, 0.1, 0.25, 0.5};
    final static FitnessFunction fitnessFunction = new SimonsFitnessFunction ();
    final static Initializer initializer = new RowCompletionInitializer ();

    static int[][] rawGrids;
    static boolean[][] fixedGrids;
    static int[][] initGrids;

    //static int[] rawGrid;
    //static int[] initGrid;
    //static boolean[] fixedGrid;
    //static final int boxSize = 3;

    public static void main (String[] args) throws Exception {
        rawGrids = new int[][]{Sudoku.readGrid("grids/easy1.txt"), Sudoku.readGrid("grids/easy2.txt"),
                Sudoku.readGrid("grids/easy3.txt"),
                Sudoku.readGrid("grids/medium1.txt"),
                Sudoku.readGrid("grids/hard1.txt"),
                Sudoku.readGrid("grids/fbfeasy1.txt")};
        fixedGrids = new boolean[rawGrids.length][rawGrids[0].length];
        initGrids = new int[rawGrids.length][rawGrids[0].length];
        for (int i = 0; i < rawGrids.length; i++) {
            Sudoku.boxSize = (int) Math.sqrt (Math.sqrt (rawGrids[i].length));
            fixedGrids[i] = Sudoku.getFixedPositions (rawGrids[i]);
            initGrids[i] = initializer.initializeGrid (fixedGrids[i], rawGrids[i], Sudoku.boxSize);
        }
        /*rawGrid = Sudoku.readGrid ("grids/easy1.txt");
        fixedGrid = Sudoku.getFixedPositions (rawGrid);
        initGrid = initializer.initializeGrid (fixedGrid, rawGrid, boxSize);
        doTests1 (hammingCrossovers, hammingMutators, populationSizes, proportionsSteadyState, mutationRates, args[0]);*/
        doTests2 (initGrids, fixedGrids, swapCrossovers, swapMutators, 5000, 2500, 0.8, args[0]);
    }

    public static void doTests2 (int[][] grids, boolean[][] fixed, Crossover[] crossovers, Mutator[] mutators,
                                 int popSize, int startOfChildren, double mutationRate, String fileName) {
        for (int grid = 0; grid < grids.length; grid++) {
            for (int crossover = 0; crossover < crossovers.length; crossover++) {
                for (int mutator = 0; mutator < mutators.length; mutator++) {
                    doTest (grids[grid], fixed[grid], crossovers[crossover], mutators[mutator],
                            popSize, startOfChildren, mutationRate, fileName);
                }
            }
        }
    }

/*    public static void doTests1 (Crossover[] crossovers, Mutator[] mutators,
                                int[] populationSizes, double[] proportionsSteadyState, double[] mutationRates,
                                String fileName) {

        for (int popIndex = 0; popIndex < populationSizes.length; popIndex++) {
            for (int steadyIndex = 0; steadyIndex < proportionsSteadyState.length; steadyIndex++) {
                for (int mutationIndex = 0; mutationIndex < mutationRates.length; mutationIndex++) {
                    for (int crossover = 0; crossover < crossovers.length; crossover++) {
                        for (int mutator = 0; mutator < mutators.length; mutator++) {
                            int popSize = populationSizes[popIndex];
                            int startOfChildren = (int) (popSize * proportionsSteadyState[steadyIndex]);
                            double mutationRate = mutationRates[mutationIndex];
                            doTest (crossovers[crossover], mutators[mutator],
                                    popSize, startOfChildren, mutationRate, fileName);
                        }
                    }
                }
            }
        }

    }       */

    public static void doTest (int[] grid, boolean[] fixed, Crossover crossover, Mutator mutator, int popSize, int startOfChildren,
                               double mutationRate, String fileName) {
        TestResults results = evolveSeveralTimes (grid, fixed, crossover, mutator,
                popSize, startOfChildren, mutationRate, numberOfTestsPerConfiguration,
                maxNumberOfEvaluationsWithoutProgress / popSize);
        try {
            BufferedWriter resultWriter = new BufferedWriter (new FileWriter (fileName, true));
            String paramString = crossover.getClass ().toString () + " " +
                    mutator.getClass ().toString () + " " + popSize + " "
                    + startOfChildren + " " + mutationRate + "\n";
            String resultString = results.numberOfSuccesses + " " +
                    results.generationBestReached.mean () +
                    " " + results.generationBestReached.sd () + " " +
                    results.bestFitness.mean () + " " +
                    results.bestFitness.sd ()+ "\n";
            System.out.print (paramString);
            System.out.print (resultString);
            resultWriter.write (paramString);
            resultWriter.write (resultString);
            resultWriter.close ();
        }
        catch (Exception e) {
            e.printStackTrace ();
        }
    }

    public static TestResults evolveSeveralTimes
            (int[] grid, boolean[] fixed, Crossover crossover, Mutator mutator, int popSize, int startOfChildren, double mutationRate,
             int numberOfTests, int maxGens) {
        int boxSize = (int) (Math.sqrt (Math.sqrt (grid.length)));
        int rowLength = boxSize * boxSize;
        final int optimumFitness = (rowLength * rowLength) * 3;
        TestResults totalResults = new TestResults ();
        for (int i = 0; i < numberOfTests; i++) {
            int[] results = evolveOnce (grid, fixed, crossover, mutator, popSize, startOfChildren, mutationRate,
                    maxGens, boxSize);
            int maxFitness = results[0];
            int whichGeneration = results[1];
            totalResults.bestFitness.add (maxFitness);
            if (maxFitness == optimumFitness) {
                totalResults.numberOfSuccesses++;
                totalResults.generationBestReached.add (whichGeneration);
            }
            System.out.print (".");
        }
        System.out.println ();
        return totalResults;
    }

    public static int[] evolveOnce (int[] grid, boolean[] fixed, Crossover crossover, Mutator mutator, int popSize,
                                    int startOfChildren, double mutationRate, int maxGens, int boxSize) {
        int[] results = new int[2]; // max fitness, generation at which this was attained
        EA ea = new NewSteadyState (fitnessFunction, mutator, crossover, popSize, startOfChildren);
        int rowLength = boxSize * boxSize;
        final int optimumFitness = (rowLength * rowLength) * 3;
        ea.initialize (fixed, grid, boxSize);
        mutator.setMutationRate (mutationRate);
        int generationsWithoutProgress = 0;
        int fitness = 0;
        int lastFitness = 0;
        int generation = 0;
        do {
            ea.oneMoreGeneration ();
            fitness = (int) ea.getMaximumFitness ();
            if (fitness > lastFitness) {
                generationsWithoutProgress = 0;
                lastFitness = fitness;
            }
            else {
                generationsWithoutProgress++;
            }
            generation++;
        } while (generationsWithoutProgress < maxGens && fitness < optimumFitness);
        results[0] = fitness;
        results[1] = generation;
        return results;
    }

    static class TestResults {
        StatisticalSummary bestFitness = new StatisticalSummary ();
        StatisticalSummary generationBestReached = new StatisticalSummary ();
        int numberOfSuccesses = 0;
    }

}
