package games.sudoku;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 10-Oct-2005
 * Time: 23:24:52
 * To change this template use File | Settings | File Templates.
 */
public class HillClimber implements EA {

    private FitnessFunction fitnessFunction;
    private Mutator mutator;
    private int[] grid = null;
    private boolean[] fixed = null;
    private double currentFitness = 0;
    private int boxSize;
    private boolean changeSinceLastGeneration = false;
    // Determines wether to move to a new genotype that has identical fitness to the parent
    //private final boolean crawlNeutralNetworks = true;


    public HillClimber (FitnessFunction fitnessFunction, Mutator mutator) {
        this.fitnessFunction = fitnessFunction;
        this.mutator = mutator;

        // And we don't need any crossover with a population of one. Of course.
    }

    public void initialize (boolean[] fixed, int[] grid, int boxSize) {
        int[] copiedGrid = new int[grid.length];
        System.arraycopy (grid, 0, copiedGrid, 0, grid.length);
        this.grid = copiedGrid;
        this.boxSize = boxSize;
        this.fixed = fixed;
        currentFitness = fitnessFunction.evaluate (grid, boxSize);
    }

    public void oneMoreGeneration () {
        mutator.mutate (fixed, grid, boxSize);
        double newFitness = fitnessFunction.evaluate (grid, boxSize);
        if (newFitness < currentFitness) {
            mutator.reverseLastMutation (fixed, grid, boxSize);
            changeSinceLastGeneration = false;
        }
        else {
            currentFitness = newFitness;
            changeSinceLastGeneration = true;
        }
    }

    public int[] bestGrid () {
        return grid;
    }

    public boolean changeSinceLastGeneration () {
        return changeSinceLastGeneration;
    }

    public int[][] getPopulation () {
        int[][] population = new int[1][1];
        population[0] = grid;
        return population;
    }

    public double getMaximumFitness () {
        return currentFitness;
    }

    public double getAverageFitness () {
        return currentFitness;
    }
}
