package games.sudoku;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 16-Oct-2005
 * Time: 20:07:01
 * To change this template use File | Settings | File Templates.
 */
public class MemeticAlgorithm implements EA {

    final FitnessFunction fitnessFunction;
    final Mutator mutator;
    final Crossover crossover;
    final int popSize = 100;
    final int maxNumberOfEvaluationsWithoutProgress = 50;

    private int boxSize;
    private int rowLength;
    private int boardSize;
    private int[][] population;
    private boolean[] fixed;
    private double[] fitnesses;
    private final HillClimber hillClimber;
    private boolean changeSinceLastGeneration;

    public MemeticAlgorithm (Crossover crossover, FitnessFunction fitnessFunction, Mutator mutator) {
        this.crossover = crossover;
        this.fitnessFunction = fitnessFunction;
        this.mutator = mutator;
        this.hillClimber = new HillClimber (fitnessFunction, mutator);
    }

    public void initialize (boolean[] fixed, int[] grid, int boxSize) {
        this.fixed = fixed;
        this.boxSize = boxSize;
        this.rowLength = boxSize * boxSize;
        this.boardSize = rowLength * rowLength;
        population = new int[popSize][boardSize];
        fitnesses = new double[popSize];
        RowPermuter rowPermuter = new RowPermuter ();
        for (int i = 0; i < popSize; i++) {
            System.arraycopy (grid, 0, population[i], 0, grid.length);
            rowPermuter.permute (fixed, grid, rowLength);
        }
    }

    public double getMaximumFitness () {
        return fitnesses[0];
    }

    public double getAverageFitness () {
        double sum = 0;
        for (int i = 0; i < fitnesses.length; i++) {
            sum += fitnesses[i];
        }
        return sum / fitnesses.length;
    }

    public void oneMoreGeneration () {
        int parent1 = (int) (Math.random () * Math.random () * popSize);
        int parent2 = (int) (Math.random () * Math.random () * popSize);
        int[] child = crossover.doCrossover (fixed, population[parent1], population[parent2], boxSize);
        hillClimber.initialize (fixed, child, boxSize);
        double childFitness = fitnessFunction.evaluate (child, boxSize);
        int evaluationsSinceLastProgress = 0;
        while (evaluationsSinceLastProgress < maxNumberOfEvaluationsWithoutProgress) {
            hillClimber.oneMoreGeneration ();
            if (hillClimber.getMaximumFitness () > childFitness) {
                childFitness = hillClimber.getMaximumFitness ();
                evaluationsSinceLastProgress = 0;
            }
            else {
                evaluationsSinceLastProgress++;
            }
        }
        population[popSize - 1] = child;
        int[] lastBest = population[0];
        for (int i = 0; i < population.length; i++) {
            fitnesses[i] = fitnessFunction.evaluate (population[i], boxSize);
        }
        sortPopulationByFitness ();
        if (lastBest == population[0]) {
            changeSinceLastGeneration = false;
        }
        else {
            changeSinceLastGeneration = true;
        }
    }

    public int[] bestGrid () {
        return population[0];
    }

    public boolean changeSinceLastGeneration () {
        return changeSinceLastGeneration;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public int[][] getPopulation () {
        return population;
    }

    public void sortPopulationByFitness () {
        // insertionsort in descending order
        for (int i = 0; i < population.length; i++) {
            for (int j = i + 1; j < population.length; j++) {
                if (fitnesses[i] < fitnesses[j]) {
                    double cache = fitnesses[i];
                    fitnesses[i] = fitnesses[j];
                    fitnesses[j] = cache;
                    int[] gcache = population[i];
                    population[i] = population[j];
                    population[j] = gcache;
                }
            }
        }
    }
}
