package games.sudoku;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 19-Oct-2005
 * Time: 16:41:23
 * To change this template use File | Settings | File Templates.
 */
public class NewSteadyState implements EA {

    final FitnessFunction fitnessFunction;
    final Mutator mutator;
    final Crossover crossover;
    final int popSize;
    final int startOfChildren;

    private int boxSize;
    private int rowLength;
    private int boardSize;

    private int[][] population;
    private boolean[] fixed;
    private double[] fitnesses;
    private boolean changeSinceLastGeneration;

    public NewSteadyState (FitnessFunction fitnessFunction, Mutator mutator, Crossover crossover,
                           int popSize, int startOfChildren) {
        this.crossover = crossover;
        this.fitnessFunction = fitnessFunction;
        this.mutator = mutator;
        this.popSize = popSize;
        this.startOfChildren = startOfChildren;
    }

    public void initialize (boolean[] fixed, int[] grid, int boxSize) {
        this.fixed = fixed;
        this.boxSize = boxSize;
        this.rowLength = boxSize * boxSize;
        this.boardSize = rowLength * rowLength;
        population = new int[popSize][boardSize];
        fitnesses = new double[popSize];
        RowPermuter rowPermuter = new RowPermuter ();
        for (int i = 0; i < popSize; i++) {
            System.arraycopy (grid, 0, population[i], 0, grid.length);
            rowPermuter.permute (fixed, grid, rowLength);
        }
    }

    public double getMaximumFitness () {
        return fitnesses[0];
    }

    public double getAverageFitness () {
        double sum = 0;
        for (int i = 0; i < fitnesses.length; i++) {
            sum += fitnesses[i];
        }
        return sum / fitnesses.length;
    }

    public void oneMoreGeneration () {
        for (int i = startOfChildren; i < popSize; i++) {
            int parent1pos = (int) (Math.random () * startOfChildren);
            int parent2pos = (int) (Math.random () * startOfChildren);
            population[i] = crossover.doCrossover (fixed, population[parent1pos], population[parent2pos], boxSize);
            mutator.mutate (fixed, population[i], boxSize);
        }
        for (int i = 0; i < population.length; i++) {
            fitnesses[i] = fitnessFunction.evaluate (population[i], boxSize);
        }
        int[] lastBest = population[0];
        sortPopulationByFitness ();
        if (lastBest == population[0]) {
            changeSinceLastGeneration = false;
        }
        else {
            changeSinceLastGeneration = true;
        }
    }

    public int[] bestGrid () {
        return population[0];
    }

    public boolean changeSinceLastGeneration () {
        return changeSinceLastGeneration;
    }

    public int[][] getPopulation () {
        return population;
    }

    public void sortPopulationByFitness () {
        // insertionsort in descending order
        for (int i = 0; i < population.length; i++) {
            for (int j = i + 1; j < population.length; j++) {
                if (fitnesses[i] < fitnesses[j]) {
                    double cache = fitnesses[i];
                    fitnesses[i] = fitnesses[j];
                    fitnesses[j] = cache;
                    int[] gcache = population[i];
                    population[i] = population[j];
                    population[j] = gcache;
                }
            }
        }
    }

}
