
package games.sudoku;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.Random;

/**
 * Created by IntelliJ IDEA.
 * User: sml
 * Date: 05-Oct-2005
 * Time: 09:29:24
 * To change this template use File | Settings | File Templates.
 */
public class SimonsSudoku {

    static int size = 9;
    static int subGrid = 3;
    int[] match;

    public static void main(String[] args) throws Exception {
        int[] grid = readGrid("grids/solved1.txt");
        printGrid(System.out, grid);
        SimonsSudoku s = new SimonsSudoku();
        System.out.println("Fitness = " + s.fitness(grid));
        // swap some elements
        Random r = new Random();
        for (int i=0; i<1000; i++) {
            int a = r.nextInt(grid.length);
            int b = r.nextInt(grid.length);
            swap(grid, a, b);
            System.out.println(i + "\t " + s.fitness(grid));
        }
    }

    public SimonsSudoku() {
        match = new int[size + 1];
    }

    public static void swap(int[] grid, int i, int j) {
        int tmp = grid[i];
        grid[i] = grid[j];
        grid[j] = tmp;
    }

    public int fitness(int[] grid) {
        // check all the rows and columns
        int tot = 0;
        for (int i = 0; i < size; i++) {
            // check a column
            tot += checkLine(grid, i, size);
            // check a row
            tot += checkLine(grid, i * size, 1);
        }
        // now check the sub-grids
        for (int x = 0; x < size; x += subGrid) {
            for (int y = 0; y < size; y += subGrid) {
                tot += checkBox(grid, x, y);
            }
        }
        return tot;
    }

    public int checkLine(int[] grid, int start, int diff) {
        int tot = 0;
        resetMatch();
        for (int i = 0; i < size; i++) {
            int val = grid[start];
            if (match[val] == 0) {
                match[val] = 1;
                tot++;
            }
            start += diff;
        }
        return tot;
    }

    public int checkBox(int[] grid, int x, int y) {
        int tot = 0;
        resetMatch();
        for (int i = x; i < x + subGrid; i++) {
            for (int j = 0; j < y + subGrid; j++) {
                int val = grid[i + size * j];
                if (match[val] == 0) {
                    match[val] = 1;
                    tot++;
                }
            }
        }
        return tot;
    }

    public void resetMatch() {
        match[0] = 1;
        for (int i = 1; i < match.length; i++) {
            match[i] = 0;
        }
    }

    public static int[] readGrid(String file) throws Exception {
        int[] grid = new int[size*size];
        BufferedReader br = new BufferedReader(new FileReader(file));
        int ix = 0;
        for (int i=0; i<size; i++) {
            String line = br.readLine();
            for (int j=0; j<size; j++) {
                char ch = line.charAt(j);
                if (ch > '0' && ch <= '9') {
                    grid[ix] = ch - '0';
                } else {
                    grid[ix] = 0;
                }
                ix++;
            }
        }
        return grid;
    }

    public static void printGrid(PrintStream out, int[] grid) {
        out.println();
        for (int i=0; i<grid.length; i++) {
            out.print(grid[i]);
            if (((i+1)%size) == 0) {
                out.println();
            }
        }
        out.println();
    }
}
