package games.sudoku;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 12-Oct-2005
 * Time: 19:05:33
 * To change this template use File | Settings | File Templates.
 */
public class SimpleGA implements EA {

    private final FitnessFunction fitnessFunction;
    private final Mutator mutator;
    private final Crossover crossover;

    final static int popSize = 100;
    final static int numberOfCrossovers = popSize / 2;
    final static int numberOfMutations = popSize / 2;

    private int boxSize;
    private int rowLength;
    private int boardSize;

    private int[][] population;
    private boolean[] fixed;
    private double[] fitnesses;
    private double maximumFitness;
    private boolean changeSinceLastGeneration;

    public SimpleGA (FitnessFunction fitnessFunction, Mutator mutator, Crossover crossover) {
        this.crossover = crossover;
        this.fitnessFunction = fitnessFunction;
        this.mutator = mutator;
    }

    public void initialize (boolean[] fixed, int[] grid, int boxSize) {
        this.fixed = fixed;
        this.boxSize = boxSize;
        this.rowLength = boxSize * boxSize;
        this.boardSize = rowLength * rowLength;
        population = new int[popSize][boardSize];
        fitnesses = new double[popSize];
        double startingFitness = fitnessFunction.evaluate (grid, boxSize);
        for (int i = 0; i < population.length; i++) {
            for (int j = 0; j < boardSize; j++) {
                population[i][j] = grid[j];
            }
            fitnesses[i] = startingFitness;
        }
        maximumFitness = startingFitness;
    }

    public double getMaximumFitness () {
        return maximumFitness;
    }

    public double getAverageFitness () {
        double sum = 0;
        for (int i = 0; i < fitnesses.length; i++) {
            sum += fitnesses[i];
        }
        return sum / fitnesses.length;
    }

    public void oneMoreGeneration () {
        // tournament selection
        // compare three genotypes, replace the least fit with a crossover of the two others
        int victim;
        int parent1;
        int parent2;
        for (int i = 0; i < numberOfCrossovers; i++) {
            int first = random (), second = random (), third = random ();
            if (fitnesses[first] < fitnesses[second]) {
                if (fitnesses[first] < fitnesses[third]) {
                    victim = first;
                    parent1 = second;
                    parent2 = third;
                }
                else {
                    victim = third;
                    parent1 = first;
                    parent2 = second;
                }
            }
            else {
                if (fitnesses[second] < fitnesses[third]) {
                    victim = second;
                    parent1 = first;
                    parent2 = third;
                }
                else {
                    victim = third;
                    parent1 = first;
                    parent2 = second;
                }
            }
            population[victim] = crossover.doCrossover (fixed, population[parent1], population[parent2], boxSize);
            //System.out.println ("p " + population[victim].length);
            fitnesses[victim] = fitnessFunction.evaluate (population[victim], boxSize);
            if (fitnesses[victim] > maximumFitness)
                maximumFitness = fitnesses[victim];
        }
        // compare two genotypes, replace the least fit with a copy of the more fit
        for (int i = 0; i < numberOfMutations; i++) {
            int first = random (), second = random ();
            if (fitnesses[first] < fitnesses[second]) {
                victim = first;
                parent1 = second;
            }
            else {
                victim = second;
                parent1 = first;
            }
            System.arraycopy (population[parent1], 0, population[victim], 0, boardSize);
            mutator.mutate (fixed, population[victim], boxSize);
            fitnesses[victim] = fitnessFunction.evaluate (population[victim], boxSize);
            if (fitnesses[victim] > maximumFitness)
                maximumFitness = fitnesses[victim];
        }
    }

    public int[] bestGrid () {
        return population[0];
    }

    public boolean changeSinceLastGeneration () {
        return changeSinceLastGeneration;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public int[][] getPopulation () {
        return population;
    }

    public int random () {
        return (int) (Math.random () * population.length);
    }

}
