package games.sudoku;

import problems.tracker.IntStack;

import java.util.Random;

/**
 * Created by IntelliJ IDEA.
 * User: sml
 * Date: 05-Oct-2005
 * <p/>
 * The idea behind this class is to model all
 * the possible values that a square could take,
 * given the hard constraints.
 * <p/>
 * Each square can then perform things like a random
 * mutation to a different value
 */
public class SmartSquare {



    public static void main(String[] args) {
        // test the ways in which a SmartSquare can
        // be used

        // first create a default one
        SmartSquare ss = new SmartSquare();
        ss.remove(7);
        ss.remove(3);
        ss.remove(4);
        ss.setStack();
        System.out.println(ss);

    }

    static Random r = new Random();

    public SmartSquare() {
        taken = new boolean[size + 1];
        possible = new IntStack(size);
        setStack();
    }

    public SmartSquare(int val) {
        // this constructor is used for a fixed square
        this();
        fix(val);
    }

    static int size = SimonsSudoku.size;
    int val; // the current value of the square
    boolean valid;

    boolean[] taken;
    IntStack possible;

    // possible.

    public int nValues() {
        return possible.size();
    }

    public void fix(int val) {
        // possible;
        possible.reset();
        for (int i = 0; i < taken.length; i++) {
            taken[i] = true;
        }
        this.val = val;
    }

    public void set(int val) {
        this.val = val;
    }

    public void remove(int val) {
        int orig = possible.size();
        if (!taken[val]) {
            taken[val] = true;
            setStack();
            // System.out.println(orig + " -> " + possible.size());
            if (possible.size() < 2) {
                // System.out.println("Fixing: " + this);
            }
        }
    }

    public int get() {
        if (fixed()) {
            return val;
        } else {
            return 0;
        }
    }

    public void setStack() {
        possible.reset();
        for (int i = 1; i < size + 1; i++) {
            if (!taken[i]) {
                possible.push(i);
            }
        }
        if (possible.size() == 1) {
            val = possible.peek(0);
        }
    }

    public boolean mutable() {
        return possible != null && possible.size() > 1;
    }

    public boolean fixed() {
        return !mutable();
    }

    public int mutate() {
        if (mutable()) {
            val = possible.peek(r.nextInt(possible.size()));
            // System.out.println("Rand val = " + val);
        }
        if (val == 0) {
            throw new RuntimeException("Zero val");
        }
        return val;
    }

    public String toString() {
        if (fixed()) {
            return "Fixed to: " + val;
        }
        if (possible.size() == 1) {
            setStack();
            return "Should fix to: " + possible.peek(0);
        }
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < possible.size(); i++) {
            sb.append(possible.peek(i));
            sb.append(" ");
        }
        return sb.toString();
    }
}
