package games.sudoku;

import problems.tracker.IntStack;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.Random;

import utilities.StatisticalSummary;

/**
 * This accounts for the basic constraints.
 * <p/>
 * Next - run the random hill climber on it
 */
public class SmartSudoku {

    public final boolean useExponentialProbabilityMutation;
    public double mutationProbability = 0.8;

    static int maxFitness = 243;
    static int size = 9;
    static int subGrid = 3;
    int[] match;
    SmartSquare[] grid;
    int[] g;
    IntStack mutable;
    static Random r = new Random();
    int mutSquare, previous;
    SimonsSudoku s;
    static int noImproveLimit = 500;

    public static void main(String[] args) throws Exception {
        SmartSudoku ss = readGrid("grids/hard3.txt");
        ss.getGrid();
        SudokuView.display("Initial: " + ss.fitness(), ss.getGrid());
        // printGrid(System.out, ss.grid);
        Sudoku.printGrid(System.out, ss.getGrid());
        for (int i = 0; i < 10; i++) {
            ss.propagate();
            System.out.println("Have Propagated");
            Sudoku.printGrid(System.out, ss.getGrid());
        }
        SudokuView.display("Propagated: " + ss.fitness(), ss.getGrid());
        printGrid(System.out, ss.grid);

        for (int i = 0; i < 10; i++) {
            // ss.mutate();
        }

        int maxEvals = 100000;
        ss.randHillBitFlip(maxEvals);
        SudokuView.display("Fitness: " + ss.fitness(), ss.g);


    }

    public static int doTest (int maxEvalsWithoutProgress, int[] grid, boolean useExponentialProbabilityMutation) {
        SmartSudoku ss = new SmartSudoku (useExponentialProbabilityMutation);
        for (int i = 0; i < grid.length; i++) {
            if (grid[i] == 0) {

                ss.grid[i] = new SmartSquare ();
                //System.out.print (grid[i]);
            }
            else {

                ss.grid[i] = new SmartSquare (grid[i]);
                //System.out.print (grid[i]);
            }
        }
        //System.out.println ();
        //ss.printTheGrid ();
        for (int i = 0; i < 10; i++) {
            ss.propagate();
        }
        //ss.printTheGrid ();
        return ss.hillClimb (maxEvalsWithoutProgress);
    }

    public SmartSudoku(boolean useExponentialProbabilityMutation) {
        this.useExponentialProbabilityMutation = useExponentialProbabilityMutation;
        grid = new SmartSquare[size * size];
        g = new int[size * size];
        for (int i = 0; i < grid.length; i++) {
            // need setting up?
        }
        s = new SimonsSudoku();
    }

    public int fitness() {
        return s.fitness(g);
    }

    public void printTheGrid () {
        for (int i = 0; i < grid.length; i++) {
            System.out.print (grid[i].fixed () ? "x" : " ");
        }
        System.out.println ();
    }

    public int hillClimb (int maxEvalsWithoutProgress) {
        g = randGrid();
        int maxFitness = (size * size) * 3;
        int f = s.fitness(g);
        if (f == maxFitness) {
            System.out.println ("Fitness " + f + " immediately!");
            return f;
        }

        int noImproves = 0;
        int gen = 0;
        do {
            if (useExponentialProbabilityMutation) {
                while (Math.random () < mutationProbability) {
                    mutate ();
                }
            }
            else {
                mutate ();
            }
            int cur = s.fitness(g);
            if (cur == maxFitness) {
                System.out.println ("Optimum at: " + gen);
                return cur;
            }
            if (cur > f) {
                f = cur;
                noImproves = 0;
                //System.out.println(gen + "\t " + f);
            } else {
                noImproves++;
                if (cur < f) {
                    undo();
                }
            }
            gen++;
        } while (noImproves < maxEvalsWithoutProgress);
        System.out.println ("Fitness " + f + " at iteration " + gen);
        return f;
    }

    public int randHillBitFlip(int maxEvals) {
        g = randGrid();
        SudokuView.display("Rand Fill: " + fitness(), g);
        int f = s.fitness(g);
        int noImproves = 0;
        for (int i = 0; i < maxEvals; i++) {
            // make the change at p
            mutate();
            int cur = s.fitness(g);
            if (cur == maxFitness) {
                System.out.println("Optimum at: " + i);
                return cur;
            }
            if (cur > f) {
                f = cur;
                noImproves = 0;
                System.out.println(i + "\t " + f);
            } else if (cur <= f) {
                // revert to old value
                noImproves++;
                if (noImproves > noImproveLimit) {
                    noImproves = 0;
                    g = randGrid();
                    f = s.fitness(g);
                    System.out.println(i + " -> Restarted: " + f);
                } else if (cur < f) {
                    undo();
                }
            }
        }
        return f;
    }

    public void mutate() {
        if (mutable == null) {
            mutable = new IntStack(size * size);
            //System.out.println (mutable.size ());
            for (int i = 0; i < grid.length; i++) {
                if (grid[i].mutable()) {
                    mutable.push(i);
                    //System.out.print (".");
                }
                else {
                    //System.out.print (grid[i]);
                }
            }
           // System.out.println (mutable.size ());
        }
        // choose a mutable square at random
        mutSquare = mutable.peek(r.nextInt(mutable.size()));
        previous = g[mutSquare];
        // now make a mutation
        g[mutSquare] = grid[mutSquare].mutate();
    }

    public void undo() {
        g[mutSquare] = previous;
        grid[mutSquare].set(previous);
    }


    public int[] getGrid() {
        int nFixed = 0;
        for (int i = 0; i < grid.length; i++) {
            g[i] = grid[i].get();
            if (g[i] > 0) {
                nFixed++;
            }
        }
        System.out.println("nFixed = " + nFixed);
        return g;
    }

    public int[] randGrid() {
        for (int i = 0; i < grid.length; i++) {
            grid[i].mutate();
            g[i] = grid[i].val;
            if (g[i] == 0) {
                throw new RuntimeException(("Zero!!!!"));
            }
        }
        return g;
    }

    public void propagate() {
        for (int i = 0; i < grid.length; i++) {
            if (grid[i].fixed()) {
                // propagate the constraints
                propagate(grid[i].val, i);
            }
        }
    }

    public void propagate(int val, int p) {
        int x = p % size;
        int y = p / size;
        // propagate through the rows and cols
        for (int i = 0; i < size; i++) {
            grid[x + i * size].remove(val);
            grid[i + y * size].remove(val);
        }

        // now find it's box, and propagate around that
        int bx = subGrid * (x / subGrid);
        int by = subGrid * (y / subGrid);

        // now see what else we need...
        for (int i = bx; i < bx + subGrid; i++) {
            for (int j = by; j < by + subGrid; j++) {
                grid[i + j * size].remove(val);
            }
        }
    }

    public void backProp() {
        // for each constraint group, need to find all
        // the numbers
        // for each number in each constraint group,
        // find the possible places where this may occur

        // each one has a fix

    }


    public void resetMatch() {
        match[0] = 1;
        for (int i = 1; i < match.length; i++) {
            match[i] = 0;
        }
    }

    public static SmartSudoku readGrid(String file) throws Exception {
        SmartSudoku ss = new SmartSudoku(false);
        BufferedReader br = new BufferedReader(new FileReader(file));
        int ix = 0;
        for (int i = 0; i < size; i++) {
            String line = br.readLine();
            for (int j = 0; j < size; j++) {
                char ch = line.charAt(j);
                if (ch > '0' && ch <= '9') {
                    // fixed
                    ss.grid[ix] = new SmartSquare(ch - '0');
                } else {
                    // choice
                    ss.grid[ix] = new SmartSquare();
                }
                ix++;
            }
        }
        return ss;
    }

    public static void printGrid(PrintStream out, SmartSquare[] grid) {
        out.println();
        for (int i = 0; i < grid.length; i++) {
            out.println(i + "\t " + grid[i]);
//            if (((i+1)%size) == 0) {
//                out.println();
//            }
        }
        out.println();
        out.flush();
    }
}
