package games.sudoku;

import java.io.*;
import java.util.List;
import java.util.LinkedList;
import java.util.StringTokenizer;
import java.awt.*;

/**
 * Created by IntelliJ IDEA.
 * User: sml
 * Date: 05-Oct-2005
 * Time: 09:29:24
 * To change this template use File | Settings | File Templates.
 */
public class Sudoku {

    static int boxSize;
    static int rowLength;
    final static List<double[]> listory = new LinkedList<double[]> ();

    public static void main(String[] args) throws Exception {

        Mutator mutator;
        mutator = new NonMutator ();
        //mutator = new ExponentialProbabilityRowSwapMutator ();
        //mutator = new RowSwapMutator ();
        //Mutator mutator = new NaiveMutator ();
        //Mutator mutator = new UniformSwapMutator ();
        Crossover crossover;
        //crossover = new TwoPointCrossover ();
        //crossover = new UniformCrossover ();
        //crossover = new NoCrossover ();
        crossover = new PMXCrossover ();
        //crossover = new UniformSwapCrossover ();
        //crossover = new OneCycleCrossover ();
        //crossover = new MultiCycleCrossover ();
        //Initializer initializer = new NaiveInitializer ();
        Initializer initializer = new RowCompletionInitializer ();
        FitnessFunction fitnessFunction = new SimonsFitnessFunction ();
        int[] originalGrid = readGrid ("grids/easy1.txt");
        int[] solutionGrid = readGrid ("grids/solutiontoeasy1.txt");
        //int[] originalGrid = new int[rowLength * rowLength];
        //int[] originalGrid = makeGrid ();
        System.out.println ("Original fitness: " + fitnessFunction.evaluate (originalGrid, boxSize));
        int[] grid = new int[originalGrid.length];

        System.arraycopy (originalGrid, 0, grid, 0, originalGrid.length);
        printGrid (System.out, grid);
        System.out.println ("Fixed elements");
        boolean[] fixed = getFixedPositions (grid);
        //boolean[] fixed = new boolean[grid.length];
        printFixedGrid (System.out, fixed);
        initializer.initializeGrid (fixed, grid, boxSize);
        printGrid (System.out, grid);
        System.out.println ("Fitness after initialization: " + fitnessFunction.evaluate (grid, boxSize));

        EA ea;
        //ea = new HillClimber (fitnessFunction, mutator);
        //ea = new SimpleGA (fitnessFunction, mutator, crossover);
        //ea = new SteadyState (fitnessFunction, mutator, crossover);
        ea = new NewSteadyState (fitnessFunction, mutator, crossover, 5000, 250);
        //ea = new MemeticAlgorithm (crossover, fitnessFunction, mutator);
        ea.initialize (fixed, grid, boxSize);
        double fitness = 0;
        double average = 0;
        double distance = 0;
        final int maxFitness = (rowLength * rowLength) * 3;
        int generation = 0;
        int numberOfChangesSinceLastGeneration = 0;
        int lastSampleDistance = Integer.MAX_VALUE;
        int newSampleDistance;
        final int maxGen = 400;
        do {
            ea.oneMoreGeneration ();
            if (ea.changeSinceLastGeneration ()) {
                numberOfChangesSinceLastGeneration++;
            }
            newSampleDistance = PopulationMeasures.sampledAverageMeanDistance (ea.getPopulation (), boxSize);
            fitness = ea.getMaximumFitness ();
            average = ea.getAverageFitness ();
            distance = swapDistance (ea.bestGrid (), solutionGrid, rowLength);
            listory.add (new double[]{fitness, average, distance, newSampleDistance});
            if ((ea.getMaximumFitness () > fitness) || (newSampleDistance < lastSampleDistance)) {
                lastSampleDistance = newSampleDistance;


                System.out.println ("Generation " + generation + " fitness " + fitness +  " average " + average + " distance " +
                        distance + " changes " +
                        numberOfChangesSinceLastGeneration + " sample " +
                        newSampleDistance);
                        //+ " thorough " + PopulationMeasures.thoroughAverageMeanDistance (ea.getPopulation (), boxSize));

                numberOfChangesSinceLastGeneration = 0;
                //+ " last mutated " + ((NaiveMutator) mutator).lastMutated + " to " + ((NaiveMutator) mutator).lastValue);
                //printGrid (System.out,  grid);
            }
            generation++;
        } while (fitness < maxFitness && generation < maxGen);
        if (fitness < maxFitness)
            System.out.println ("No, we did not get there!");
        printGrid (System.out, originalGrid);
        printGrid (System.out, solutionGrid);
        printGrid (System.out, ea.bestGrid ());
        BufferedWriter bw = new BufferedWriter (new FileWriter ("history.txt", false));
        for (double[] values : listory) {
            String string = "";
            for (int i = 0; i < values.length; i++) {
                string += values[i] + " ";
            }
            bw.write (string + "\n");
        }
        bw.close ();
    }



    public static void swap(int[] grid, int i, int j) {
        int tmp = grid[i];
        grid[i] = grid[j];
        grid[j] = tmp;
    }


    public static boolean[] getFixedPositions (int[] grid) {
        boolean fixed[] = new boolean[grid.length];
        for (int i = 0; i < grid.length; i++) {
            if (grid[i] != 0)
                fixed[i] = true;
        }
        return fixed;
    }

    public static int[] readGrid(String file) throws Exception {
        BufferedReader br = new BufferedReader(new FileReader(file));
        String line = br.readLine ();
        StringTokenizer st = new StringTokenizer (line, " ");
        boxSize = Integer.parseInt (st.nextToken ());
        rowLength = boxSize * boxSize;
        System.out.println ("Box size set to " + boxSize);
        System.out.println ("Loaded grid of type " + st.nextToken ());
        int[] grid = new int[rowLength*rowLength];
        int ix = 0;
        for (int i=0; i<rowLength; i++) {
            line = br.readLine();
            st = new StringTokenizer (line, " ");
            int j = 0;
            while (st.hasMoreTokens ()) {
                int value = Integer.parseInt (st.nextToken ());
                grid[i * rowLength + j] = value;
                j++;
            }
        }
        return grid;
    }

    public static void printGrid(PrintStream out, int[] grid) {
        out.println();
        for (int i=0; i<grid.length; i++) {
            out.print(grid[i] > 9 ? grid[i] + " " : grid[i] + "  ");
            if (((i+1)%rowLength) == 0) {
                out.println();
            }
        }
        out.println();
    }

    public static void printFixedGrid(PrintStream out, boolean[] fixed) {
        out.println();
        for (int i=0; i<fixed.length; i++) {
            out.print(fixed[i] ? "*  " : "   ");
            if (((i+1)%rowLength) == 0) {
                out.println();
            }
        }
        out.println();
    }

    public static int[] makeGrid() {
        // returns a grid with the n of each number 1..n
        int[] x = new int[rowLength*rowLength];
        int ix = 0;
        for (int i=1; i<= rowLength; i++) {
            for (int j=1; j<= rowLength; j++) {
                x[ix++] = j;
            }
        }
        return x;
    }

    public static void printTotalOccurences (int[] grid) {
        int[] occurences = new int[rowLength + 1];
        for (int i = 0; i < grid.length; i++) {
            occurences[grid[i]]++;
        }
        System.out.print ("Total Occurences of numbers: ");
        for (int i = 0; i < occurences.length; i++) {
            System.out.print (" " + occurences[i] + " ");
        }
        System.out.println ();
    }

    public static void printRowOccurences (int[] grid, int boxSize) {
        int rowLength = boxSize * boxSize;
        System.out.print ("Occurences of numbers per row: ");
        for (int i = 0; i < rowLength; i++) {
            printOccurencesForOneRow (grid, i * rowLength, rowLength);
        }
    }

    public static void printOccurencesForOneRow (int[] grid, int offset, int rowLength) {
        int[] occurences = new int[rowLength + 1];
        for (int i = offset; i < offset + rowLength; i++) {
            occurences[grid[i]]++;
        }
        for (int i = 0; i < occurences.length; i++) {
            System.out.print (" " + occurences[i] + " ");
        }
        System.out.println ();
    }

    public static int hammingDistance (int[] grid1, int[] grid2) {
        int distance = 0;
        for (int i = 0; i < grid1.length; i++) {
            if (grid1[i] != grid2[i]) {
                distance++;
            }
        }
        return distance;
    }

    public static int swapDistance (int[] grid1, int[] grid2original, int rowLength) {
        int[] grid2 = new int[grid2original.length];
        System.arraycopy (grid2original, 0, grid2, 0, grid2.length);
        int total = 0;
        for (int i = 0; i < rowLength; i++) {
            total += swapDistance (grid1, grid2, rowLength * i, rowLength);
        }
        return total;
    }

    public static int swapDistance (int[] grid1, int[] grid2, int offset, int rowLength) {
        // This method is destructive. grid2 will be reordered. Resistance is futile.
        int swaps = 0;
        for (int i = offset; i < offset + rowLength; i++) {
            if (grid1[i] != grid2[i]) {
                // we need to do a swap in grid2
                // find the place in grid2 where the value we're looking for is
                int j;
                for (j = offset; j < offset + rowLength; j++) {
                    if (grid2[j] == grid1[i]) {
                        break;
                    }
                }
                swap (grid2, i, j);
                swaps++;
            }
        }
        return swaps;
    }

}
