package games.sudoku;

import utilities.StatisticalSummary;

import java.io.BufferedWriter;
import java.io.FileWriter;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 08-Dec-2005
 * Time: 05:58:35
 * To change this template use File | Settings | File Templates.
 */
public class DoHCStats {

    final static Mutator[] mutators = {new NonMutator (), new NaiveMutator (), new UniformSwapMutator (),
                                       new ExponentialProbabilityNaiveMutator (), new RowSwapMutator (),
                                       new ExponentialProbabilityRowSwapMutator ()};
    final static FitnessFunction fitnessFunction = new SimonsFitnessFunction ();
    final static Initializer initializer = new RowCompletionInitializer ();
    //static int[] rawGrid;
    //static int[] initGrid;
    //static boolean[] fixedGrid;
    static final int boxSize = 3;
    static final int maxNumberOfEvaluationsWithoutProgress = 100000;
    final static int numberOfTestsPerConfiguration = 30;
    //final static double[] mutationRates = {0, 0.2, 0.4, 0.6, 0.8};
    static int[][] rawGrids;
    static boolean[][] fixedGrids;
    static int[][] initGrids;

    public static void main (String[] args) throws Exception {
        rawGrids = new int[][]{Sudoku.readGrid("grids/easy1.txt"), Sudoku.readGrid("grids/easy2.txt"),
                               Sudoku.readGrid("grids/easy3.txt"),
                               Sudoku.readGrid("grids/medium1.txt"),
                               Sudoku.readGrid("grids/hard1.txt")};
        //Sudoku.readGrid("grids/fbfeasy1.txt")};
        fixedGrids = new boolean[rawGrids.length][rawGrids[0].length];
       // initGrids = new int[rawGrids.length][rawGrids[0].length];
        for (int i = 0; i < rawGrids.length; i++) {
            Sudoku.boxSize = (int) Math.sqrt (Math.sqrt (rawGrids[i].length));
            fixedGrids[i] = Sudoku.getFixedPositions (rawGrids[i]);
            //initGrids[i] = initializer.initializeGrid (fixedGrids[i], rawGrids[i], Sudoku.boxSize);
        }
        //rawGrid = Sudoku.readGrid ("grids/easy1.txt");
        //fixedGrid = Sudoku.getFixedPositions (rawGrid);
        //initGrid = initializer.initializeGrid (fixedGrid, rawGrid, boxSize);
        //doTests (mutators, new double[]{0.8}, args[0]);
        doSmartBoxTests (args[0], numberOfTestsPerConfiguration);
    }

    public static void doSmartBoxTests (String fileName, int numberOfTestsPerConfiguration) {
        for (int grid = 0; grid < rawGrids.length; grid++) {
            StatisticalSummary ss = new StatisticalSummary ();
            for (int i = 0; i < numberOfTestsPerConfiguration; i++) {
                int fitness = SmartSudoku.doTest (maxNumberOfEvaluationsWithoutProgress, rawGrids[grid], true);
                ss.add (fitness);
            }
            System.out.println (ss);
        }
    }

    public static void doTests (Mutator[] mutators, double[] mutationRates, String fileName) {
        for (int mutationIndex = 0; mutationIndex < mutationRates.length; mutationIndex++) {
            for (int grid = 0; grid < initGrids.length; grid++) {
                for (int mutator = 0; mutator < mutators.length; mutator++) {
                    double mutationRate = mutationRates[mutationIndex];
                    doTest (mutationRate, mutators[mutator], initGrids[grid], fixedGrids[grid], grid, fileName);
                }
            }
        }
    }

    public static void doTest (double mutationRate, Mutator mutator, int[] initGrid, boolean[] fixedGrid, int whichGrid, String fileName) {
        try {
            TestResults results = evolveSeveralTimes (mutator,
                    mutationRate, numberOfTestsPerConfiguration, initGrid, fixedGrid);
            BufferedWriter resultWriter = new BufferedWriter (new FileWriter (fileName, true));
            String paramString =
                    mutator.getClass ().toString () + " grid-" + whichGrid +  " " + mutationRate + "\n";
            String resultString = results.numberOfSuccesses + " " +
                    results.generationBestReached.mean () +
                    " " + results.generationBestReached.sd () + " " +
                    results.bestFitness.mean () + " " +
                    results.bestFitness.sd ()+ "\n";
            System.out.print (paramString);
            System.out.print (resultString);
            resultWriter.write (paramString);
            resultWriter.write (resultString);
            resultWriter.close ();
        }
        catch (Exception e) {
            e.printStackTrace ();
        }
    }

    public static TestResults evolveSeveralTimes
            (Mutator mutator, double mutationRate, int numberOfTests, int[] initGrid, boolean[] fixedGrid) {
        TestResults totalResults = new TestResults ();
        for (int i = 0; i < numberOfTests; i++) {
            int[] results = evolveOnce (mutator, mutationRate, initGrid, fixedGrid);
            int maxFitness = results[0];
            int whichGeneration = results[1];
            totalResults.bestFitness.add (maxFitness);
            if (maxFitness == 243) {
                totalResults.numberOfSuccesses++;
                totalResults.generationBestReached.add (whichGeneration);
            }
            System.out.print (".");
        }
        System.out.println ();
        return totalResults;
    }


    public static int[] evolveOnce (Mutator mutator, double mutationRate, int[] initGrid, boolean[] fixedGrid) {
        int[] results = new int[2]; // max fitness, generation at which this was attained
        EA ea = new HillClimber (fitnessFunction, mutator);
        ea.initialize (fixedGrid, initGrid, boxSize);
        mutator.setMutationRate (mutationRate);
        int generationsWithoutProgress = 0;
        int fitness = 0;
        int lastFitness = 0;
        int generation = 0;
        int maxFitness = 243;
        do {
            ea.oneMoreGeneration ();
            fitness = (int) ea.getMaximumFitness ();
            if (fitness > lastFitness) {
                generationsWithoutProgress = 0;
                lastFitness = fitness;
            }
            else {
                generationsWithoutProgress++;
            }
            generation++;
        } while (generationsWithoutProgress < maxNumberOfEvaluationsWithoutProgress && fitness < maxFitness);
        results[0] = fitness;
        results[1] = generation;
        return results;
    }

    static class TestResults {
        StatisticalSummary bestFitness = new StatisticalSummary ();
        StatisticalSummary generationBestReached = new StatisticalSummary ();
        int numberOfSuccesses = 0;
    }

}
